import datetime
from decimal import Decimal


def get_annual_rate_of_return(starting_value: Decimal, ending_value: Decimal, number_of_years: Decimal) -> Decimal:
    a: Decimal = ending_value ** (1 / number_of_years)
    b: Decimal = (1 / starting_value) ** (1 / number_of_years)
    return (a * b) - Decimal("1")


def get_year_frac(starting_date: datetime.datetime, ending_date: datetime.datetime) -> Decimal:
    number_of_whole_years: Decimal = Decimal(ending_date.year - starting_date.year)

    new_starting_date: datetime.datetime
    try:
        new_starting_date = datetime.datetime(starting_date.year + int(number_of_whole_years), starting_date.month, starting_date.day)
    except ValueError as e:
        if starting_date.month == 2 and starting_date.day == 29:
            new_starting_date = datetime.datetime(int(starting_date.year + number_of_whole_years), 3, 1)
        else:
            raise e

    difference_in_days: Decimal = Decimal((ending_date - new_starting_date).days)
    return number_of_whole_years + get_year_frac_from_number_of_days(difference_in_days, ending_date.year)


def get_number_of_days_from_year_frac(number_of_fractional_years: Decimal, year: int) -> Decimal:
    day_count_basis: Decimal = Decimal("365") if year % 4 != 0 else Decimal("366")
    return Decimal(int((number_of_fractional_years * day_count_basis)))


def get_year_frac_from_number_of_days(number_of_days: Decimal, year: int) -> Decimal:
    day_count_basis: Decimal = Decimal("365") if year % 4 != 0 else Decimal("366")
    return number_of_days / day_count_basis


def add_years_to_date(date: datetime.datetime, number_of_years: Decimal) -> datetime:
    number_of_whole_years: Decimal = Decimal(int(number_of_years))
    number_of_fractional_years: Decimal = number_of_years - number_of_whole_years

    future_date: datetime.datetime
    try:
        future_date = date.replace(year=int(date.year + number_of_whole_years))
    except ValueError as e:
        if date.month == 2 and date.day == 29:
            future_date = datetime.datetime(int(date.year + number_of_whole_years), 3, 1)
        else:
            raise e

    return future_date + datetime.timedelta(days=float(get_number_of_days_from_year_frac(number_of_fractional_years, future_date.year)))



import datetime
import time
from dataclasses import dataclass
from decimal import Decimal
from typing import Dict, List, Optional

from kih_api.investment_analysis import calculations
from kih_api.investment_analysis.models import Statistics, HistoricalStockPriceData, NumberOfYearsTillPerpetualProfit


@dataclass
class StockAnalysis:
    stock_symbol: str
    analysis_start_date: datetime.datetime
    analysis_end_date: datetime.datetime
    historical_data_list: List[HistoricalStockPriceData]
    investment_time_frame: Decimal
    absolute_return: Optional[Statistics]
    annualized_return: Optional[Statistics]
    number_of_years_till_perpetual_profit: NumberOfYearsTillPerpetualProfit

    def __init__(self, stock_symbol: str, analysis_start_date: str, analysis_end_date: str, investment_time_frame: float):
        self.stock_symbol = stock_symbol
        self.analysis_start_date = datetime.datetime.strptime(analysis_start_date, "%Y-%m-%d")
        self.analysis_end_date = datetime.datetime.strptime(analysis_end_date, "%Y-%m-%d")
        self.investment_time_frame = Decimal(str(investment_time_frame))
        self.get_historical_data()
        self.calculate()

    def get_historical_data(self) -> None:
        all_historical_data_list: List[HistoricalStockPriceData] = HistoricalStockPriceData.get(self.stock_symbol)
        filtered_historical_data_list: List[HistoricalStockPriceData] = list(filter(lambda data: self.analysis_start_date <= data.timestamp <= self.analysis_end_date, all_historical_data_list))
        sorted_historical_data_list: List[HistoricalStockPriceData] = sorted(filtered_historical_data_list, key=lambda data: data.timestamp)
        self.historical_data_list = sorted_historical_data_list

    def calculate(self) -> None:
        self.calculate_returns()
        self.calculate_number_of_years_till_no_loss()

    def calculate_returns(self) -> None:
        absolute_return_list: List[Dict[datetime.datetime, Decimal]] = []
        annualized_return_list: List[Dict[datetime.datetime, Decimal]] = []

        for starting_historical_data in self.historical_data_list:
            end_date: datetime.datetime = calculations.add_years_to_date(starting_historical_data.timestamp, self.investment_time_frame)
            ending_historical_data: HistoricalStockPriceData = HistoricalStockPriceData.get_from_date(end_date, self.historical_data_list)

            if ending_historical_data is None:
                break

            year_frac: Decimal = calculations.get_year_frac(starting_historical_data.timestamp, ending_historical_data.timestamp)

            absolute_return: Decimal = calculations.get_annual_rate_of_return(starting_historical_data.close, ending_historical_data.close, Decimal("1"))
            annualized_return: Decimal = calculations.get_annual_rate_of_return(starting_historical_data.close, ending_historical_data.close, year_frac)

            absolute_return_list.append({starting_historical_data.timestamp: absolute_return})
            annualized_return_list.append({starting_historical_data.timestamp: annualized_return})

        if len(absolute_return_list) > 0:
            self.absolute_return = Statistics.get([data[next(iter(data))] for data in absolute_return_list])
            self.absolute_return.data = absolute_return_list

        if len(annualized_return_list) > 0:
            self.annualized_return = Statistics.get([data[next(iter(data))] for data in annualized_return_list])
            self.annualized_return.data = annualized_return_list

    def calculate_number_of_years_till_no_loss(self) -> None:
        start_time = time.time_ns()
        self.number_of_years_till_perpetual_profit = NumberOfYearsTillPerpetualProfit.get_from_historical_stock_price_data_list(self.historical_data_list, True)
        print(f"Time taken: {(time.time_ns() - start_time) / 1_000_000_000 / 60} minutes")


test = StockAnalysis("TQQQ", "2010-01-01", "2023-01-01", 3)



from dataclasses import dataclass
from datetime import datetime, timedelta
from decimal import Decimal
from typing import List, Union

from kih_api.ibkr.helper import HistoricalDataHelper
from kih_api.ibkr.models import HistoricalData, IBKR
from kih_api.investment_analysis import calculations
from kih_api.investment_analysis.custom_leveraged_trading.helper import InvestmentAnalysisHelper
from kih_api.investment_analysis.models import Statistics


@dataclass
class HistoricalReturn:
    date: datetime
    number_of_shares_bought: Decimal
    net_liquidity: Decimal
    close: Decimal


@dataclass
class InvestmentAnalysis:
    symbol: str
    starting_date: datetime
    ending_date: datetime
    raw_historical_data_list: List[HistoricalData]
    historical_data_list: List[HistoricalData]
    historical_returns_list: List[HistoricalReturn]
    profit: Decimal
    annual_rate_of_return: Decimal
    number_of_days_till_no_loss: Decimal
    lowest_net_liquidity: Decimal

    def __init__(self, symbol: str, starting_balance: Decimal, daily_leverage: Decimal, starting_date: Union[datetime, str] = None, ending_date: Union[datetime, str] = None, raw_historical_data_list: List[HistoricalData] = None):
        starting_date = datetime.utcfromtimestamp(1) if starting_date is None else datetime.strptime(starting_date, '%Y-%m-%d') if isinstance(starting_date, str) else starting_date
        ending_date = datetime.today() if ending_date is None else datetime.strptime(ending_date, '%Y-%m-%d') if isinstance(ending_date, str) else ending_date

        self.symbol = symbol
        self.raw_historical_data_list = IBKR.get_historical_data(symbol) if raw_historical_data_list is None else raw_historical_data_list
        self.historical_data_list = HistoricalDataHelper.filter_historical_data_list(starting_date, ending_date, self.raw_historical_data_list)
        self.starting_date = self.historical_data_list[0].timestamp
        self.ending_date = self.historical_data_list[-1].timestamp
        self.historical_returns_list = InvestmentAnalysisHelper.get_leveraged_historical_data_list(starting_balance, daily_leverage, self.historical_data_list)
        self.profit = self.historical_returns_list[-1].net_liquidity - starting_balance
        self.annual_rate_of_return = calculations.get_annual_rate_of_return(starting_balance, self.historical_returns_list[-1].net_liquidity, calculations.get_year_frac(self.starting_date, self.ending_date))
        self.number_of_days_till_no_loss, self.lowest_net_liquidity = InvestmentAnalysisHelper.get_number_of_days_till_no_loss(self.historical_returns_list)


@dataclass
class InvestmentAnalysisStatistics:
    profit: Statistics
    annual_rate_of_return: Statistics
    number_of_days_till_no_loss: Statistics
    lowest_net_liquidity: Statistics

    def __init__(self, investment_analysis_list: List[InvestmentAnalysis]):
        self.profit = Statistics([data.profit for data in investment_analysis_list])
        self.annual_rate_of_return = Statistics([data.annual_rate_of_return for data in investment_analysis_list])
        self.number_of_days_till_no_loss = Statistics([data.number_of_days_till_no_loss for data in investment_analysis_list])
        self.lowest_net_liquidity = Statistics([data.lowest_net_liquidity for data in investment_analysis_list])


@dataclass
class InvestmentAnalysisSimulation:
    symbol: str
    investment: Decimal
    leverage: Decimal
    starting_date: datetime
    ending_date: datetime
    number_of_years_of_investment: Decimal
    investment_analysis_list: List[InvestmentAnalysis]
    raw_historical_data_list: List[HistoricalData]
    investment_analysis_statistics: InvestmentAnalysisStatistics

    def __init__(self, symbol: str, investment: Decimal, leverage: Decimal, starting_date: Union[datetime, str], ending_date: Union[datetime, str], number_of_years_of_investment: Decimal,
                 raw_historical_data_list: List[HistoricalData] = None):
        self.starting_date = datetime.utcfromtimestamp(1) if starting_date is None else datetime.strptime(starting_date, '%Y-%m-%d') if isinstance(starting_date, str) else starting_date
        self.ending_date = datetime.today() if ending_date is None else datetime.strptime(ending_date, '%Y-%m-%d') if isinstance(ending_date, str) else ending_date

        self.symbol = symbol
        self.investment = investment
        self.leverage = leverage
        self.number_of_years_of_investment = Decimal(str(number_of_years_of_investment))
        self.raw_historical_data_list = IBKR.get_historical_data(self.symbol) if raw_historical_data_list is None else raw_historical_data_list
        self.get_investment_analysis_list()
        self.investment_analysis_statistics = InvestmentAnalysisStatistics(self.investment_analysis_list)

    def get_investment_analysis_list(self) -> None:
        self.investment_analysis_list = []
        investment_starting_date = max(self.starting_date, self.raw_historical_data_list[0].timestamp)
        next_investment_starting_date: datetime = investment_starting_date
        next_investment_ending_date = calculations.add_years_to_date(next_investment_starting_date, self.number_of_years_of_investment)

        while next_investment_ending_date <= min(self.ending_date, self.raw_historical_data_list[-1].timestamp):
            self.investment_analysis_list.append(
                InvestmentAnalysis(self.symbol, self.investment, self.leverage, starting_date=next_investment_starting_date, ending_date=next_investment_ending_date, raw_historical_data_list=self.raw_historical_data_list))

            next_investment_starting_date = next_investment_starting_date + timedelta(days=1)
            next_investment_ending_date = calculations.add_years_to_date(next_investment_starting_date, self.number_of_years_of_investment)



import datetime
import statistics
from dataclasses import dataclass, asdict
from decimal import Decimal
from typing import List, Any, Optional, cast, Dict

from mongoengine import StringField, FloatField, DateTimeField, EmbeddedDocumentField, EmbeddedDocument, ListField

import kih_api.ibkr.models
from kih_api.database.models import DatabaseDocument
from kih_api.investment_analysis import calculations


@dataclass
class Statistics(EmbeddedDocument):
    maximum: Decimal = FloatField(required=True)
    median: Decimal = FloatField(required=True)
    average: Decimal = FloatField(required=True)
    minimum: Decimal = FloatField(required=True)
    standard_deviation: Decimal = FloatField(required=True)

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.maximum = Decimal(str(self.maximum))
        self.median = Decimal(str(self.median))
        self.average = Decimal(str(self.average))
        self.minimum = Decimal(str(self.minimum))
        self.standard_deviation = Decimal(str(self.standard_deviation))

    @staticmethod
    def get(data: List[Decimal]) -> "Statistics":
        return Statistics(
            **{"maximum": Decimal(str(max(data))),
               "median": Decimal(str(statistics.median(data))),
               "average": Decimal(str(statistics.mean(data))),
               "minimum": Decimal(str(min(data))),
               "standard_deviation": Decimal(str(statistics.stdev(data)))}
        )


class HistoricalStockPriceData(DatabaseDocument):
    stock_symbol: str = StringField(required=True)
    open: Decimal = FloatField()
    high: Decimal = FloatField()
    low: Decimal = FloatField()
    close: Decimal = FloatField(required=True)
    volume: Decimal = FloatField()
    timestamp: datetime.datetime = DateTimeField(required=True)

    meta = {"indexes": ["stock_symbol", "timestamp"]}

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.open = Decimal(str(self.open))
        self.high = Decimal(str(self.high))
        self.low = Decimal(str(self.low))
        self.close = Decimal(str(self.close))
        self.volume = Decimal(str(self.volume))

    @staticmethod
    def get(stock_symbol: str, from_date: datetime.datetime = datetime.datetime.now().replace(year=(datetime.datetime.now().year - 10))) -> List["HistoricalStockPriceData"]:
        historical_data_list: List[HistoricalStockPriceData] = sorted(cast(List[HistoricalStockPriceData], HistoricalStockPriceData.objects(stock_symbol=stock_symbol, timestamp__gte=from_date)), key=lambda data: data.timestamp)

        if len(historical_data_list) > 0 and historical_data_list[-1].timestamp.date() == datetime.datetime.now().date():
            return historical_data_list

        updated_historical_data_list: List["HistoricalStockPriceData"] = HistoricalStockPriceData.get_updated_data(stock_symbol, historical_data_list[-1].timestamp if len(historical_data_list) > 0 else from_date)
        return historical_data_list + updated_historical_data_list

    @staticmethod
    def get_updated_data(stock_symbol: str, from_date: datetime.datetime) -> List["HistoricalStockPriceData"]:
        new_historical_data_list: List[HistoricalStockPriceData] = []
        if kih_api.ibkr.models.IBKR.is_connection_successful():
            new_historical_data_list = HistoricalStockPriceData.get_from_ibkr(stock_symbol, from_date)

        if len(new_historical_data_list) > 1:
            HistoricalStockPriceData.objects.insert(new_historical_data_list)

        return new_historical_data_list

    @staticmethod
    def get_from_ibkr(stock_symbol: str, from_date: datetime.datetime = datetime.datetime.now().replace(year=(datetime.datetime.now().year - 10))) -> List["HistoricalStockPriceData"]:
        historical_data_list: List[HistoricalStockPriceData] = []
        ibkr_historical_data_list: List[
            kih_api.ibkr.models.HistoricalData] = kih_api.ibkr.models.IBKR.get_historical_data(stock_symbol)
        for ibkr_historical_data in ibkr_historical_data_list:
            if ibkr_historical_data.timestamp > from_date:
                ibkr_historical_data_dict: Dict[str, Any] = asdict(ibkr_historical_data)
                ibkr_historical_data_dict.pop("average")
                ibkr_historical_data_dict["stock_symbol"] = stock_symbol
                historical_data_list.append(HistoricalStockPriceData(**ibkr_historical_data_dict))

        return historical_data_list

    @staticmethod
    def get_from_date(date: datetime.datetime, historical_data_list: List["HistoricalStockPriceData"], is_get_next_date_if_unavailable: bool = True) -> Optional["HistoricalStockPriceData"]:
        filtered_historical_data_list: List[HistoricalStockPriceData] = sorted(list(filter(lambda data: data.timestamp >= date, historical_data_list)), key=lambda data: data.timestamp)
        if len(filtered_historical_data_list) == 0 or (not is_get_next_date_if_unavailable and filtered_historical_data_list[0] != date):
            return None
        return filtered_historical_data_list[0]


class PerpetualProfitDate(EmbeddedDocument):
    investment_date: datetime.datetime = DateTimeField(required=True)
    perpetual_profit_date: datetime.datetime = DateTimeField(required=True)
    number_of_years: Decimal = FloatField(required=True)

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.number_of_years = Decimal(str(self.number_of_years))


class NumberOfYearsTillPerpetualProfit(DatabaseDocument):
    stock_symbol: str = StringField(primary_key=True)
    data: List[PerpetualProfitDate] = ListField(EmbeddedDocumentField(PerpetualProfitDate))
    statistics: Statistics = EmbeddedDocumentField(Statistics)

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_from_historical_stock_price_data_list(historical_stock_price_data: List[HistoricalStockPriceData], update_database: bool = False) -> "NumberOfYearsTillPerpetualProfit":
        perpetual_profit_date_list: List[PerpetualProfitDate] = []
        if len(historical_stock_price_data) == 0:
            raise Exception("No market data")

        for historical_data in historical_stock_price_data:
            profit_historical_data_list: List[HistoricalStockPriceData] = list(filter(lambda data: data.close >= historical_data.close and data.timestamp > historical_data.timestamp, historical_stock_price_data))

            date_of_no_loss: datetime.datetime = None

            for profit_historical_data in profit_historical_data_list:
                future_loss_historical_data_list: List[HistoricalStockPriceData] = list(filter(lambda data: data.close < historical_data.close and data.timestamp > profit_historical_data.timestamp, historical_stock_price_data))
                if len(future_loss_historical_data_list) == 0:
                    date_of_no_loss = profit_historical_data.timestamp
                    break

            if date_of_no_loss is None:
                continue

            year_frac: Decimal = calculations.get_year_frac(historical_data.timestamp, date_of_no_loss)
            perpetual_profit_date_list.append(PerpetualProfitDate(investment_date=historical_data.timestamp, perpetual_profit_date=date_of_no_loss, number_of_years=year_frac))

        number_of_years_till_perpetual_profit: NumberOfYearsTillPerpetualProfit = NumberOfYearsTillPerpetualProfit(
            stock_symbol=historical_stock_price_data[0].stock_symbol,
            data=perpetual_profit_date_list,
            statistics=Statistics.get([perpetual_profit_date.number_of_years for perpetual_profit_date in perpetual_profit_date_list])
        )

        if update_database:
            number_of_years_till_perpetual_profit.save()

        return number_of_years_till_perpetual_profit



import typing
from datetime import datetime
from decimal import Decimal
from typing import List
from typing import Tuple

from kih_api.ibkr.models import HistoricalData

if typing.TYPE_CHECKING:
    from kih_api.investment_analysis.custom_leveraged_trading.models import HistoricalData, HistoricalReturn


class InvestmentAnalysisHelper:

    @staticmethod
    def get_leveraged_historical_data_list(starting_balance: Decimal, leverage: Decimal, historical_data_list: List["HistoricalData"]) -> List["HistoricalReturn"]:
        from kih_api.investment_analysis.custom_leveraged_trading.models import HistoricalReturn
        historical_returns_list: List[HistoricalReturn] = []

        number_of_shares_bought: Decimal = Decimal(0)
        loan_amount: Decimal = Decimal(0)
        for historical_data in historical_data_list:
            if historical_data == historical_data_list[0]:
                cash_balance: Decimal = starting_balance
            else:
                cash_balance = (number_of_shares_bought * historical_data.close) - loan_amount if number_of_shares_bought != Decimal(0) else cash_balance

            if cash_balance <= Decimal(0):
                historical_returns_list.append(HistoricalReturn(historical_data.timestamp, Decimal(0), Decimal(0), historical_data.close))
                return historical_returns_list

            number_of_shares_bought = Decimal(int((cash_balance * leverage) / historical_data.close))
            loan_amount = (number_of_shares_bought * historical_data.close) - cash_balance if (number_of_shares_bought * historical_data.close) > cash_balance else Decimal(0)
            net_liquidity = (number_of_shares_bought * historical_data.close) - loan_amount if (number_of_shares_bought * historical_data.close) > cash_balance else cash_balance

            historical_returns_list.append(HistoricalReturn(historical_data.timestamp, number_of_shares_bought, net_liquidity, historical_data.close))
        return historical_returns_list

    @staticmethod
    def get_number_of_days_till_no_loss(historical_returns_list: List["HistoricalReturn"]) -> Tuple[Decimal, Decimal]:
        starting_date: datetime = historical_returns_list[0].date
        lowest_net_liquidity: Decimal = historical_returns_list[0].net_liquidity
        number_of_days_till_no_loss: Decimal = Decimal(0)
        for historical_return in historical_returns_list:
            number_of_days_till_no_loss = Decimal((historical_return.date - starting_date).days) if historical_return.net_liquidity < lowest_net_liquidity else number_of_days_till_no_loss
            lowest_net_liquidity = historical_return.net_liquidity if historical_return.net_liquidity < lowest_net_liquidity else lowest_net_liquidity
        return number_of_days_till_no_loss, lowest_net_liquidity




from decimal import Decimal

from kih_api.global_common import Currency
from kih_api.wise.models import IntraAccountTransfer, ReserveAccount, ProfileTypes, CashAccount

monthly_expenses_reserve_account: ReserveAccount = ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(ProfileTypes.PERSONAL, Currency.NZD, "Monthly Expenses", False)
nzd_account: CashAccount = CashAccount.get_by_profile_type_and_currency(ProfileTypes.PERSONAL, Currency.NZD)
test = IntraAccountTransfer.execute(Decimal("50"), monthly_expenses_reserve_account, nzd_account, ProfileTypes.PERSONAL)
pass


from __future__ import annotations

import enum
from dataclasses import dataclass
from decimal import Decimal
from typing import List, Optional, Union

import kih_api
from kih_api import global_common, communication
from kih_api.finance_database.exceptions import InsufficientFundsException, AccountForCurrencyNotFoundException
from kih_api.http_requests import ClientErrorException
from kih_api.logger import logger
from kih_api.wise import wise_models
from kih_api.wise.exceptions import MultipleUserProfilesWithSameTypeException, MultipleRecipientsWithSameAccountNumberException, TransferringMoneyToNonSelfOwnedAccountsException, ReserveAccountNotFoundException


class ProfileTypes(enum.Enum):
    PERSONAL: str = "personal"
    BUSINESS: str = "business"


class AccountType(enum.Enum):
    CashAccount: str = "STANDARD"
    ReserveAccount: str = "SAVINGS"


@dataclass
class UserProfile:
    id: int
    type: ProfileTypes
    first_name: str
    last_name: str
    date_of_birth: str

    @classmethod
    def get_all(cls) -> List["UserProfile"]:
        user_profiles_list: List[UserProfile] = []

        for wise_user_profile in wise_models.UserProfiles.call():
            user_profiles_list.append(UserProfile(wise_user_profile.id, global_common.get_enum_from_value(wise_user_profile.type, ProfileTypes), wise_user_profile.details.firstName, wise_user_profile.details.lastName, wise_user_profile.details.dateOfBirth))

        return user_profiles_list

    @classmethod
    def get_by_profile_type(cls, profile_type: ProfileTypes) -> "UserProfile":
        return_user_profile: UserProfile = None

        for user_profile in UserProfile.get_all():
            if user_profile.type == profile_type:
                if return_user_profile is None:
                    return_user_profile = user_profile
                else:
                    raise MultipleUserProfilesWithSameTypeException()
        return return_user_profile


@dataclass
class Account:
    id: int
    currency: global_common.Currency
    balance: Decimal
    type: AccountType
    name: str
    profile_type: ProfileTypes

    def __init__(self, wise_account: wise_models.Account):
        self.id = wise_account.id
        self.currency = global_common.get_enum_from_value(wise_account.currency, global_common.Currency)
        self.balance = Decimal(str(wise_account.cashAmount.value))
        self.type = global_common.get_enum_from_value(wise_account.type, AccountType)
        self.profile_type = global_common.get_enum_from_value(wise_account.type, AccountType)

    @classmethod
    def get_all_by_profile_type(cls, profile_type: ProfileTypes) -> List["CashAccount" | "ReserveAccount"]:
        user_profile: UserProfile = UserProfile.get_by_profile_type(profile_type)
        accounts_list: List["CashAccount" | "ReserveAccount"] = []

        for wise_account in wise_models.Account.call(user_profile.id):
            account_type: AccountType = global_common.get_enum_from_value(wise_account.type, AccountType)
            if account_type == AccountType.CashAccount:
                accounts_list.append(CashAccount(wise_account))
            elif account_type == AccountType.ReserveAccount:
                accounts_list.append(ReserveAccount(wise_account))

        return accounts_list

    @classmethod
    def get_by_profile_type_and_currency(cls, profile_type: ProfileTypes, currency: global_common.Currency) -> "CashAccount" | "ReserveAccount":
        all_accounts_list: List["CashAccount" | "ReserveAccount"] = Account.get_all_by_profile_type(profile_type)
        for account in all_accounts_list:
            if account.currency == currency:
                return account
        raise AccountForCurrencyNotFoundException()


@dataclass
class ReserveAccount(Account):
    name: str

    def __init__(self, wise_account: wise_models.Account):
        super().__init__(wise_account)
        self.name = wise_account.name

    @classmethod
    def get_all_by_profile_type(cls, profile_type: ProfileTypes) -> List["ReserveAccount"]:  # type: ignore
        all_accounts: List["CashAccount" | "ReserveAccount"] = super().get_all_by_profile_type(profile_type)
        all_cash_accounts: List[ReserveAccount] = []

        for account in all_accounts:
            if isinstance(account, ReserveAccount):
                all_cash_accounts.append(account)
        return all_cash_accounts

    @classmethod
    def get_by_profile_type_and_currency(cls, profile_type: ProfileTypes, currency: global_common.Currency) -> List["ReserveAccount"]:  # type: ignore
        all_accounts_list: List[ReserveAccount] = ReserveAccount.get_all_by_profile_type(profile_type)
        reserve_account_list: List[ReserveAccount] = []
        for account in all_accounts_list:
            if account.currency == currency:
                reserve_account_list.append(account)
        return reserve_account_list

    @classmethod
    def get_reserve_account_by_profile_type_currency_and_name(cls, profile_type: ProfileTypes, currency: global_common.Currency, name: str, create_if_unavailable: bool = False) -> "ReserveAccount":
        all_reserve_accounts_list: List[ReserveAccount] = ReserveAccount.get_by_profile_type_and_currency(profile_type, currency)
        for reserve_account in all_reserve_accounts_list:
            if reserve_account.name == name:
                return reserve_account
        if create_if_unavailable:
            return ReserveAccount.create_reserve_account(name, currency, profile_type, False)
        else:
            raise ReserveAccountNotFoundException()

    @classmethod
    def create_reserve_account(cls, name: str, currency: global_common.Currency, profile_type: ProfileTypes, check_if_available_before_creation: bool = True) -> "ReserveAccount":
        if check_if_available_before_creation:
            try:
                return ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(profile_type, currency, name)
            except ReserveAccountNotFoundException:
                pass

        wise_models.Account.create_reserve_account(name, currency, UserProfile.get_by_profile_type(profile_type).id)
        return ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(profile_type, currency, name)


@dataclass
class CashAccount(Account):

    def __init__(self, wise_account: wise_models.Account):
        super().__init__(wise_account)

    @classmethod
    def get_all_by_profile_type(cls, profile_type: ProfileTypes) -> List["CashAccount"]:  # type: ignore
        all_accounts: List["CashAccount" | "ReserveAccount"] = super().get_all_by_profile_type(profile_type)
        all_cash_accounts: List[CashAccount] = []

        for account in all_accounts:
            if isinstance(account, CashAccount):
                all_cash_accounts.append(account)
        return all_cash_accounts

    @classmethod
    def get_by_profile_type_and_currency(cls, profile_type: ProfileTypes, currency: global_common.Currency) -> "CashAccount":
        all_accounts_list: List[CashAccount] = CashAccount.get_all_by_profile_type(profile_type)
        for account in all_accounts_list:
            if account.currency == currency:
                return account
        raise AccountForCurrencyNotFoundException()


@dataclass
class ExchangeRate:
    exchange_rate: Decimal
    from_currency: global_common.Currency
    to_currency: global_common.Currency

    @classmethod
    def get(cls, from_currency: global_common.Currency, to_currency: global_common.Currency) -> "ExchangeRate":
        wise_exchange_rate: wise_models.ExchangeRate = wise_models.ExchangeRate.call(from_currency.value, to_currency.value)
        return ExchangeRate(Decimal(str(wise_exchange_rate.rate)), from_currency, to_currency)


@dataclass
class Recipient:
    account_id: int
    profile_id: int
    name: str
    currency: global_common.Currency
    is_active: bool
    is_self_owned: bool
    account_number: str
    swift_code: str
    bank_name: str
    branch_name: str
    iban: str
    bic: str

    @classmethod
    def get_all_by_profile_type(cls, profile_type: ProfileTypes) -> List["Recipient"]:
        user_profile: UserProfile = UserProfile.get_by_profile_type(profile_type)
        recipient_list: List[Recipient] = []

        for wise_recipient in wise_models.Recipient.call(user_profile.id):
            recipient_list.append(Recipient(wise_recipient.id, wise_recipient.profile, wise_recipient.accountHolderName, global_common.get_enum_from_value(wise_recipient.currency, global_common.Currency), wise_recipient.active, wise_recipient.ownedByCustomer, wise_recipient.details.accountNumber, wise_recipient.details.swiftCode, wise_recipient.details.bankName, wise_recipient.details.branchName, wise_recipient.details.swiftCode, wise_recipient.details.bic))

        return recipient_list

    @classmethod
    def get_by_account_number_and_profile_type(cls, account_number: str, profile_type: ProfileTypes) -> "Recipient":
        return_recipient: Recipient = None
        for recipient in Recipient.get_all_by_profile_type(profile_type):
            if recipient.account_number == account_number:
                if return_recipient is None:
                    return_recipient = recipient
                else:
                    raise MultipleRecipientsWithSameAccountNumberException()

        return return_recipient


@dataclass
class Transfer:
    profile_id: int
    from_currency: global_common.Currency
    to_currency: global_common.Currency
    from_amount: Decimal
    to_amount: Decimal
    recipient: Recipient
    exchange_rate: ExchangeRate
    is_successful: bool
    error_message: Optional[str]
    error_code: Optional[str]

    def __init__(self, user_profile: UserProfile, recipient: Recipient, wise_transfer: wise_models.Transfer, wise_fund: wise_models.Fund):
        self.profile_id = user_profile.id
        self.from_currency = global_common.get_enum_from_value(wise_transfer.sourceCurrency, global_common.Currency)
        self.to_currency = global_common.get_enum_from_value(wise_transfer.targetCurrency, global_common.Currency)
        self.from_amount = Decimal(str(wise_transfer.sourceValue))
        self.to_amount = Decimal(str(wise_transfer.targetValue))
        self.recipient = recipient
        self.exchange_rate = ExchangeRate(Decimal(str(wise_transfer.rate)), self.from_currency, self.to_currency)
        self.is_successful = wise_fund.status == "COMPLETED"
        self.error_message = wise_fund.errorMessage
        self.error_code = wise_fund.errorCode

    @classmethod
    def execute(cls, receiving_amount: Decimal, from_currency: global_common.Currency, to_currency: global_common.Currency, recipient_account_number: str, reference: str, profile_type: ProfileTypes) -> "Transfer":
        recipient: Recipient = Recipient.get_by_account_number_and_profile_type(recipient_account_number, profile_type)

        if not recipient.is_self_owned:
            raise TransferringMoneyToNonSelfOwnedAccountsException()

        try:
            if from_currency == to_currency:
                cash_account: CashAccount = CashAccount.get_by_profile_type_and_currency(profile_type, from_currency)
                if cash_account.balance < receiving_amount:
                    raise InsufficientFundsException(f"Insufficient funds"
                                                     f"\nRequired amount: {from_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}"
                                                     f"\nAccount Balance: {from_currency.value} {global_common.get_formatted_string_from_decimal(cash_account.balance)}"
                                                     f"\nShort of: {from_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount - cash_account.balance)}")

            user_profile: UserProfile = UserProfile.get_by_profile_type(profile_type)
            wise_quote: wise_models.Quote = wise_models.Quote.call(user_profile.id, from_currency.value, to_currency.value, float(receiving_amount))
            wise_transfer: wise_models.Transfer = wise_models.Transfer.call(recipient.account_id, wise_quote.id, reference)
            wise_fund: wise_models.Fund = wise_models.Fund.call(user_profile.id, wise_transfer.id)
            transfer: Transfer = Transfer(user_profile, recipient, wise_transfer, wise_fund)

            if transfer.is_successful:
                communication.telegram.send_message(kih_api.communication.telegram.constants.telegram_channel_username,
                                                    f"<u><b>Money transferred</b></u>"
                                                    f"\n\nAmount: <i>{to_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                    f"\nTo: <i>{recipient.name}</i>"
                                                    f"\nReference: <i>{reference}</i>", True)
            else:
                raise ClientErrorException(transfer.error_message)

            return transfer

        except ClientErrorException as e:
            logger.error(str(e))
            communication.telegram.send_message(kih_api.communication.telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nTo: <i>{recipient.name}</i>"
                                                f"\nReason: <i>{str(e)}</i>"
                                                f"\nReference: <i>{reference}</i>", True)
        except InsufficientFundsException as e:
            logger.error(str(e))
            communication.telegram.send_message(kih_api.communication.telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nTo: <i>{recipient.name}</i>"
                                                f"\nReference: <i>{reference}</i>"
                                                f"\n\nReason: <i>{str(e)}</i>", True)
            raise InsufficientFundsException(str(e))

        return None


@dataclass
class IntraAccountTransfer:
    from_currency: global_common.Currency
    from_amount: Decimal
    from_account: Account
    to_currency: global_common.Currency
    to_amount: Decimal
    to_account: Account
    exchange_rate: ExchangeRate
    is_successful: bool

    def __init__(self, from_account: Account, to_account: Account, intra_account_transfer: wise_models.IntraAccountTransfer):
        self.from_currency = global_common.get_enum_from_value(intra_account_transfer.sourceAmount.currency, global_common.Currency)
        self.from_amount = Decimal(str(intra_account_transfer.sourceAmount.value))
        self.from_account = from_account
        self.to_account = to_account
        self.to_currency = global_common.get_enum_from_value(intra_account_transfer.targetAmount.currency, global_common.Currency)
        self.to_amount = Decimal(str(intra_account_transfer.targetAmount.value))
        self.exchange_rate = ExchangeRate(Decimal(str(intra_account_transfer.rate)), self.from_currency, self.to_currency)
        self.is_successful = intra_account_transfer.state == "COMPLETED"

    # TODO: Implement withdrawal from reserve accounts
    @classmethod
    def execute(cls, receiving_amount: Decimal, from_account: Union[CashAccount, ReserveAccount], to_account: Union[CashAccount, ReserveAccount], profile_type: ProfileTypes) -> "IntraAccountTransfer":
        user_profile: UserProfile = UserProfile.get_by_profile_type(profile_type)
        intra_account_transfer: IntraAccountTransfer = None
        wise_intra_account_transfer: wise_models.IntraAccountTransfer = None

        try:
            if isinstance(to_account, CashAccount):
                if isinstance(from_account, ReserveAccount):
                    wise_intra_account_transfer = wise_models.IntraAccountTransfer.call(user_profile.id, from_account.id, to_account.id, float(receiving_amount), None, to_account.currency.value)
                else:
                    wise_quote: wise_models.Quote = wise_models.Quote.call(user_profile.id, from_account.currency.value, to_account.currency.value, float(receiving_amount))
                    wise_intra_account_transfer = wise_models.IntraAccountTransfer.call(user_profile.id, from_account.id, to_account.id, float(receiving_amount), wise_quote.id if isinstance(from_account, CashAccount) else None, None)

                intra_account_transfer = IntraAccountTransfer(from_account, to_account, wise_intra_account_transfer)

                if intra_account_transfer.is_successful:
                    communication.telegram.send_message(
                        kih_api.communication.telegram.constants.telegram_channel_username,
                                                        f"<u><b>Intra account money transferred</b></u>"
                                                        f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                        f"\nFrom: <i>{from_account.name if isinstance(from_account, ReserveAccount) else to_account.currency.value}</i>"
                                                        f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else ''} ({to_account.currency.value})</i>", True)

            elif isinstance(to_account, ReserveAccount):
                if from_account.currency != to_account.currency:
                    to_cash_account: CashAccount = CashAccount.get_by_profile_type_and_currency(profile_type, to_account.currency)
                    IntraAccountTransfer.execute(receiving_amount, from_account.currency, to_cash_account, profile_type)
                    from_account = to_cash_account

                if from_account.balance < receiving_amount:
                    raise InsufficientFundsException(f"Insufficient funds"
                                                     f"\nRequired amount: {from_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}"
                                                     f"\nAccount Balance: {from_account.currency.value} {global_common.get_formatted_string_from_decimal(from_account.balance)}"
                                                     f"\nShort of: {from_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount - from_account.balance)}")

                wise_intra_account_transfer = wise_models.IntraAccountTransfer.call(user_profile.id, from_account.id, to_account.id, float(receiving_amount), None, to_account.currency.value)
                intra_account_transfer = IntraAccountTransfer(from_account, to_account, wise_intra_account_transfer)

                if intra_account_transfer.is_successful:
                    communication.telegram.send_message(
                        kih_api.communication.telegram.constants.telegram_channel_username,
                                                        f"<u><b>Intra account money transferred</b></u>"
                                                        f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                        f"\nFrom: <i>{to_account.currency.value}</i>"
                                                        f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else ''} ({to_account.currency.value})</i>", True)
            return intra_account_transfer

        except ClientErrorException as e:
            logger.error(str(e))
            communication.telegram.send_message(kih_api.communication.telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Intra account money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nFrom: <i>{to_account.currency.value}</i>"
                                                f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else None} ({to_account.currency.value})</i>"
                                                f"\nReason: <i>{str(e)}</i>", True)
        except InsufficientFundsException as e:
            logger.error(str(e))
            communication.telegram.send_message(kih_api.communication.telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Intra account money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nFrom: <i>{to_account.currency.value}</i>"
                                                f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else to_account.currency.value} ({to_account.currency.value})</i>"
                                                f"\nReason: <i>{str(e)}</i>", True)
            raise InsufficientFundsException(str(e))

        return None
